from permission_rules.app_settings import PERMISSION_RULES_SETTINGS
from permission_rules.connect import get_redis_connect


def clear_cache(chunk_size: int = 100) -> bool:
    if not PERMISSION_RULES_SETTINGS["use_redis"]:
        return False

    r = get_redis_connect()
    prefix = PERMISSION_RULES_SETTINGS["prefix"]

    cursor = 0
    ns_keys = prefix + "*"
    while cursor != 0:
        cursor, keys = r.scan(cursor=cursor, match=ns_keys, count=chunk_size)
        if keys:
            r.delete(*keys)

    return True



from typing import List, Type

from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework import viewsets

from drf_yasg.utils import swagger_auto_schema

from permission_rules.serializers import ViewSetPermissionsSerializer
from permission_rules.services.permissions_getter import PermissionsGetter


class PermissionActionMixinMeta(type):
    __inheritors__: List[Type[viewsets.GenericViewSet]] = []

    def __new__(cls, name, bases, dict_):
        class_ = type.__new__(cls, name, bases, dict_)

        if name != "PermissionsActionMixin":
            cls.__inheritors__.append(class_)  # type: ignore

        return class_


class PermissionsActionMixin(viewsets.GenericViewSet, metaclass=PermissionActionMixinMeta):
    def _get_permissions_map(self, request):
        permissions_map = PermissionsGetter.get_viewsets_permissions([self], request)
        return permissions_map[self.__class__.__name__]

    @swagger_auto_schema(method="get", responses={200: ViewSetPermissionsSerializer})
    @action(
        methods=["GET"],
        detail=False,
        url_name="permissions",
        permission_classes=(IsAuthenticated,),
    )
    def permissions(self, request, *ags, **kwargs):
        return Response(self._get_permissions_map(request))



from django.conf import settings


if hasattr(settings, "PERMISSION_RULES_SETTINGS"):
    PERMISSION_RULES_SETTINGS = settings.PERMISSION_RULES_SETTINGS
else:
    PERMISSION_RULES_SETTINGS = {
        "use_redis": False,
        "prefix": "permission_rule_",
        "redis": {"host": "localhost", "port": "6379", "db": 0, "password": ""},
        "use_file_instead_db": False,
        "permission_rules_file_path": "",
    }



import json
from typing import Dict, List, Optional, Set, Union

from permission_rules.app_settings import PERMISSION_RULES_SETTINGS
from permission_rules.models import PermissionRule


USE_FILE = PERMISSION_RULES_SETTINGS["use_file_instead_db"]
FILE_PATH = PERMISSION_RULES_SETTINGS["permission_rules_file_path"]


class PermissionRulesFromFileContainer:
    _rules: Dict[str, PermissionRule] = {}
    _loaded: bool = False

    @classmethod
    def _load(cls):
        if cls._loaded:
            return

        with open(FILE_PATH, "r") as p_file:
            rules_data = json.loads(p_file.read())

        for p_dict in rules_data:
            p_rule_data = p_dict["fields"]

            cls._rules[p_rule_data["name"]] = PermissionRule(
                name=p_rule_data["name"],
                rule=p_rule_data["rule"],
                is_active=p_rule_data["is_active"],
            )

        cls._loaded = True

    @classmethod
    def get_permission_rule(cls, name: str) -> Optional[PermissionRule]:
        cls._load()

        return cls._rules.get(name, None)

    @classmethod
    def get_permission_rules(cls, names: Union[List[str], Set[str]]) -> List[PermissionRule]:
        cls._load()

        result = []

        for name in names:
            permission_rule = cls._rules.get(name, None)

            if permission_rule:
                result.append(permission_rule)

        return result



from django.urls import include, path

from rest_framework.routers import DefaultRouter

from permission_rules.views import PermissionsViewset


router = DefaultRouter()
router.register("", PermissionsViewset, basename="permissions")

urlpatterns = [path("permissions/", include(router.urls))]



from typing import List, Optional, Set, Union

from permission_rules.app_settings import PERMISSION_RULES_SETTINGS
from permission_rules.models import PermissionRule
from permission_rules.services.permission_rules_file import PermissionRulesFromFileContainer


USE_FILE = PERMISSION_RULES_SETTINGS["use_file_instead_db"]
FILE_PATH = PERMISSION_RULES_SETTINGS["permission_rules_file_path"]


def _get_permission_rule_from_db(name: str) -> Optional[PermissionRule]:
    try:
        return PermissionRule.objects.get(name=name)
    except PermissionRule.DoesNotExist:
        return None


def get_permission_rule(name: str) -> Optional[PermissionRule]:
    if USE_FILE:
        return PermissionRulesFromFileContainer.get_permission_rule(name)
    else:
        return _get_permission_rule_from_db(name)


def _get_permission_rules_from_db(names: Union[List[str], Set[str]]) -> List[PermissionRule]:
    return list(PermissionRule.objects.filter(name__in=names).all())


def get_permission_rules(names: Union[List[str], Set[str]]) -> List[PermissionRule]:
    if USE_FILE:
        return PermissionRulesFromFileContainer.get_permission_rules(names)
    else:
        return _get_permission_rules_from_db(names)



import redis

from permission_rules.app_settings import PERMISSION_RULES_SETTINGS


HOST = PERMISSION_RULES_SETTINGS["redis"]["host"]
PORT = PERMISSION_RULES_SETTINGS["redis"]["port"]
DB = PERMISSION_RULES_SETTINGS["redis"]["db"]
PASSWORD = PERMISSION_RULES_SETTINGS["redis"]["password"]


def get_redis_connect():
    return redis.Redis(host=HOST, port=PORT, db=DB, password=PASSWORD)


__all__ = ["get_redis_connect"]



from collections import defaultdict
from typing import Any, Dict, List, Type, Union

from rest_framework.permissions import BasePermission
from rest_framework import viewsets

from permission_rules.permission import CustomAccessPolicy
from permission_rules.services.permission_rules_getter import get_permission_rules


ViewSetName = str
ActionName = str
Permissions = List[Union[BasePermission, CustomAccessPolicy]]
PermissionClasses = List[Union[Type[BasePermission], Type[CustomAccessPolicy]]]


class CachedCustomAccessPolicy(CustomAccessPolicy):
    # For compatibility with AccessPolicy
    def __new__(cls, *args, **kwargs):
        if not args and not kwargs:
            return CustomAccessPolicy(*args, **kwargs)

        return super().__new__(cls)

    def __init__(self, name: str, original_permission: CustomAccessPolicy, cached_permissions):
        self.name = name
        self.original_permission = original_permission
        self.cached_statements = cached_permissions

    def __getattr__(self, name: str) -> Any:
        return getattr(self.original_permission, name)

    def get_policy_statements(self, request, view) -> List[dict]:
        statements = self.cached_statements.get(self.name, self.original_permission.DEFAULT_STATEMENTS)
        statements += self.original_permission.ADDITIONAL_STATEMENTS
        return statements


class PermissionsGetter:
    def __init__(self, viewsets: List[viewsets.GenericViewSet], request):
        self.viewsets = viewsets
        self.request = request

    def _get_not_detail_actions(self, viewset: viewsets.GenericViewSet) -> list:
        not_detail_actions = []

        list_action = getattr(viewset, "list", None)
        if list_action:
            not_detail_actions.append(list_action)

        create_action = getattr(viewset, "create", None)
        if create_action:
            not_detail_actions.append(create_action)

        actions = viewset.get_extra_actions()
        for view_action in actions:
            if not view_action.detail and view_action.__name__ != "permissions":
                not_detail_actions.append(view_action)

        return not_detail_actions

    def _get_viewset_permission_classes(self, viewset: viewsets.GenericViewSet) -> PermissionClasses:
        permission_classes: PermissionClasses = getattr(viewset.__class__, "permission_classes", [])

        if viewset.permission_classes:
            permission_classes += viewset.permission_classes

        return permission_classes

    def _get_cached_permissions(
        self, permission_classes_map: Dict[ViewSetName, PermissionClasses]
    ) -> Dict[ViewSetName, Permissions]:
        result = defaultdict(list)
        permission_rule_names = set()

        for viewset_name in permission_classes_map:
            for permission_class in permission_classes_map[viewset_name]:
                if issubclass(permission_class, CustomAccessPolicy):
                    permission_rule_names.add(permission_class.name)

        cached_permissions = {rule.name: rule.rule for rule in get_permission_rules(permission_rule_names)}

        for viewset_name in permission_classes_map:
            for permission_class in permission_classes_map[viewset_name]:
                if issubclass(permission_class, CustomAccessPolicy):
                    permission = CachedCustomAccessPolicy(
                        permission_class.name, permission_class(), cached_permissions
                    )
                else:
                    permission = permission_class()

                result[viewset_name].append(permission)

        return result

    def _get_permissions_map(self) -> Dict[ViewSetName, Permissions]:
        permissions_classes_map: Dict[ViewSetName, PermissionClasses] = {}

        for viewset in self.viewsets:
            name = viewset.__class__.__name__
            permissions_classes_map[name] = self._get_viewset_permission_classes(viewset)

        return self._get_cached_permissions(permissions_classes_map)

    def _check_permissions(
        self,
        viewset: viewsets.GenericViewSet,
        action,
        permissions: List[Union[BasePermission, CustomAccessPolicy]],
    ) -> bool:
        for permission in permissions:
            allowed = True

            if isinstance(permission, CustomAccessPolicy):
                allowed = permission.has_permission(self.request, viewset, action=action.__name__)
            else:
                allowed = permission.has_permission(self.request, viewset)

            if not allowed:
                return False

        return True

    def _get_viewsets_permissions(self) -> Dict[ViewSetName, Dict[ActionName, bool]]:
        permissions_map = self._get_permissions_map()

        result = {}

        for viewset in self.viewsets:
            viewset_name = viewset.__class__.__name__
            viewset_permissions = {}

            not_detail_actions = self._get_not_detail_actions(viewset)

            for action in not_detail_actions:
                viewset_permissions[action.__name__] = self._check_permissions(
                    viewset, action, permissions_map[viewset_name]
                )

            result[viewset_name] = viewset_permissions

        return result

    @classmethod
    def get_viewsets_permissions(
        cls, viewsets: List[viewsets.GenericViewSet], request
    ) -> Dict[ViewSetName, Dict[ActionName, bool]]:
        getter = cls(viewsets, request)
        return getter._get_viewsets_permissions()



from rest_framework import serializers

from drf_yasg import openapi


class ViewSetPermissionsSerializer(serializers.BaseSerializer):
    class Meta:
        swagger_schema_fields = {
            "type": openapi.TYPE_OBJECT,
            "properties": {
                "$action_name": openapi.Schema(
                    title="is allowed",
                    type=openapi.TYPE_BOOLEAN,
                ),
            },
        }


class AllViewSetsPermissionsSerializer(serializers.BaseSerializer):
    class Meta:
        swagger_schema_fields = {
            "type": openapi.TYPE_OBJECT,
            "properties": {
                "$viewset_name": ViewSetPermissionsSerializer.Meta.swagger_schema_fields,
            },
        }



from typing import Dict

from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from drf_yasg.utils import swagger_auto_schema

from permission_rules.mixins import PermissionActionMixinMeta
from permission_rules.serializers import AllViewSetsPermissionsSerializer
from permission_rules.services.permissions_getter import ActionName, PermissionsGetter, ViewSetName


class PermissionsViewset(GenericViewSet):
    def _get_viewset_map(self, request) -> Dict[ViewSetName, Dict[ActionName, bool]]:
        viewsets = [vs() for vs in PermissionActionMixinMeta.__inheritors__]

        result = {}
        permissions_map = PermissionsGetter.get_viewsets_permissions(viewsets, request)

        for viewset_name in permissions_map:
            name = viewset_name.replace("ViewSet", "")
            result[name] = permissions_map[viewset_name]

        return result

    @swagger_auto_schema(method="get", responses={200: AllViewSetsPermissionsSerializer})
    @action(methods=["GET"], detail=False)
    def get(self, request, *args, **kwargs):
        return Response(self._get_viewset_map(request))



import json

from django.db import models

from model_utils.models import TimeStampedModel

from permission_rules.app_settings import PERMISSION_RULES_SETTINGS
from permission_rules.connect import get_redis_connect


PREFIX = PERMISSION_RULES_SETTINGS["prefix"]


class PermissionRule(TimeStampedModel):
    name = models.CharField(max_length=50)
    rule = models.JSONField()
    is_active = models.BooleanField(default=True)

    def __str__(self):
        return self.name

    @property
    def dump_rule(self):
        return json.dumps(self.rule)

    @property
    def key(self):
        return f"{PREFIX}{self.name}"

    @classmethod
    def send_data_to_cache(cls):
        r = get_redis_connect()
        pipe = r.pipeline()
        for obj in cls.objects.filter(is_active=True):
            pipe.set(obj.key, obj.dump_rule)

        pipe.execute()



from rest_framework import fields

from permission_rules.permission import CustomAccessPolicy


class AccessPolicyPermissionsField(fields.Field):
    default_actions = ["create", "retrieve", "update", "destroy"]

    def __init__(self, actions=None, additional_actions=None, global_only=False, object_only=False, **kwargs):
        """See class description for parameters and usage"""
        assert not (
            global_only and object_only
        ), "Both global_only and object_only cannot be set to true on a AccessPolicyPermissionsField"
        self.action_method_map = {}

        self.global_only = global_only
        self.object_only = object_only
        self.actions = self.default_actions if (actions is None) else actions
        if additional_actions is not None:
            self.actions = self.actions + additional_actions

        kwargs["source"] = "*"
        kwargs["read_only"] = True

        self.permission = None

        super().__init__(**kwargs)

    def to_representation(self, value):
        """
        Calls the developer defined permission methods
        (both global and object) and formats the results into a dictionary.
        """
        results = {}

        request = self.context["request"]
        view = self.context["view"]
        permissions = [perm for perm in view.get_permissions() if isinstance(perm, CustomAccessPolicy)]

        if permissions:
            permission = permissions[0]
            for action in self.actions:
                global_cond = permission.has_permission(request, view, action=action)
                obj_cond = permission.has_object_permission(request, view, value, action=action)
                results[action] = global_cond and obj_cond

        return results



from django.apps import AppConfig

from permission_rules.utils import clear_cache


class PermissionRulesConfig(AppConfig):
    name = "permission_rules"

    def ready(self):
        import permission_rules.signals

        clear_cache()



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from permission_rules.app_settings import PERMISSION_RULES_SETTINGS
from permission_rules.connect import get_redis_connect
from permission_rules.models import PermissionRule


USE_REDIS = PERMISSION_RULES_SETTINGS["use_redis"]


@receiver(post_save, sender=PermissionRule)
def cache_rule_signal(sender, instance, created, **kwargs):
    if USE_REDIS and instance.is_active:
        r = get_redis_connect()
        r.delete(*[instance.key])
        r.set(instance.key, instance.dump_rule)


@receiver(post_delete, sender=PermissionRule)
def remove_rule_from_cache_signal(sender, instance, using, **kwargs):
    if USE_REDIS:
        r = get_redis_connect()
        r.delete(*[instance.key])



import json
from typing import List, Optional

from rest_access_policy import AccessPolicy, AccessPolicyException

from permission_rules.app_settings import PERMISSION_RULES_SETTINGS
from permission_rules.connect import get_redis_connect
from permission_rules.services.permission_rules_getter import get_permission_rule


USE_REDIS = PERMISSION_RULES_SETTINGS["use_redis"]
USE_FILE = PERMISSION_RULES_SETTINGS["use_file_instead_db"]
PREFIX = PERMISSION_RULES_SETTINGS["prefix"]


class CustomAccessPolicy(AccessPolicy):
    name = ""
    DEFAULT_STATEMENTS = [
        {
            "action": ["*"],
            "principal": ["*"],
            "effect": "allow",
        }
    ]
    ADDITIONAL_STATEMENTS = []
    SAFE_METHODS = ("HEAD", "OPTIONS")

    @classmethod
    def scope_queryset(cls, request, qs, action: Optional[str] = None):
        return qs.none()

    def _get_statements_matching_action(self, request, action: str, statements: List[dict]):
        """
        Filter statements and return only those that match the specified
        action.
        """
        matched = []

        for statement in statements:
            if (action in statement["action"] or "*" in statement["action"]) or (
                "<safe_methods>" in statement["action"] and request.method in self.SAFE_METHODS
            ):

                matched.append(statement)

        return matched

    def get_user_group_values(self, user) -> List[str]:
        return getattr(user, "user_group_values", [])

    def _get_policy_statements_from_db(self) -> List[dict]:
        permission_rule = get_permission_rule(self.name)

        if permission_rule:
            return permission_rule.rule

        return self.DEFAULT_STATEMENTS

    def _get_policy_statements_from_redis(self) -> List[dict]:
        key = f"{PREFIX}{self.name}"

        r = get_redis_connect()
        statements_raw = r.get(key)

        if statements_raw is None:
            statements = self._get_policy_statements_from_db()
            r.set(key, json.dumps(statements))
        else:
            statements = json.loads(statements_raw)

        return statements + self.ADDITIONAL_STATEMENTS

    def get_policy_statements(self, request, view) -> List[dict]:
        if USE_REDIS and not USE_FILE:
            statements = self._get_policy_statements_from_redis()
        else:
            statements = self._get_policy_statements_from_db()

        return statements + self.ADDITIONAL_STATEMENTS

    def has_permission(self, request, view, action: Optional[str] = None) -> bool:
        if action is None:
            action = self._get_invoked_action(view)

        statements = self.get_policy_statements(request, view)
        if len(statements) == 0:
            return False

        return self._evaluate_statements(statements, request, view, action)

    def has_object_permission(self, request, view, obj, action: Optional[str] = None):
        if action is None:
            action = self._get_invoked_action(view)

        statements = self.get_policy_statements(request, view)
        if len(statements) == 0:
            return False

        return self._evaluate_statements(statements, request, view, action, obj)

    def _evaluate_statements(self, statements: List[dict], request, view, action: str, obj=None) -> bool:
        statements = self._normalize_statements(statements)
        matched = self._get_statements_matching_principal(request, statements)
        matched = self._get_statements_matching_action(request, action, matched)

        matched = self._get_statements_matching_context_conditions(request, view, action, matched, obj)

        denied = [_ for _ in matched if _["effect"] != "allow"]

        if len(matched) == 0 or len(denied) > 0:
            return False

        return True

    def _get_statements_matching_context_conditions(
        self, request, view, action: str, statements: List[dict], obj=None
    ):
        """
        Filter statements and only return those that match all of their
        custom context conditions; if no conditions are provided then
        the statement should be returned.
        """
        matched = []
        condition_name = "condition"
        if obj is not None:
            condition_name = "obj_condition"

        for statement in statements:
            if len(statement.get(condition_name, [])) == 0:
                matched.append(statement)
                continue

            fails = 0

            for condition in statement[condition_name]:
                passed = self._check_condition(condition, request, view, action, obj)

                if not passed:
                    fails += 1
                    break

            if fails == 0:
                matched.append(statement)

        return matched

    def _check_condition(self, condition: str, request, view, action: str, obj=None):
        """
        Evaluate a custom context condition; if method does not exist on
        the access policy class, then return False.
        Condition value can contain a value that is passed to method, if
        formatted as `<method_name>:<arg_value>`.
        """
        parts = condition.split(":", 1)
        method_name = parts[0]
        arg = parts[1] if len(parts) == 2 else None
        method = self._get_condition_method(method_name)

        if arg is not None:
            result = method(request, view, action, arg)
        elif obj is not None:
            result = method(request, view, action, obj)
        else:
            result = method(request, view, action)

        if type(result) is not bool:
            raise AccessPolicyException("condition '%s' must return true/false, not %s" % (condition, type(result)))

        return result


class AdminAccessPolicy(CustomAccessPolicy):
    DEFAULT_STATEMENTS = [
        {
            "action": ["*"],
            "principal": ["group:administrators"],
            "effect": "allow",
        }
    ]



from django.contrib import admin
from django.db import models

from django_json_widget.widgets import JSONEditorWidget

from permission_rules.models import PermissionRule


@admin.register(PermissionRule)
class PermissionRuleAdmin(admin.ModelAdmin):
    list_display = ["id", "name", "is_active"]
    search_fields = ["name"]
    list_filter = ["is_active"]

    formfield_overrides = {
        models.JSONField: {"widget": JSONEditorWidget},
    }
